"""
trace-repository rules.
"""
from __future__ import annotations

from enum import Enum, unique
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Type, Union

import pandera as pa
from pydantic import BaseModel
from typer import colors

FILETYPE = "geojson"
DATABASE_CSV = "database.csv"
DATABASE_CSV_SEP = ","


@unique
class ColumnNames(Enum):

    """
    Column names for database.csv.
    """

    AREA = "area"
    TRACES = "traces"
    THEMATIC = "thematic"
    SCALE = "scale"
    AREA_SHAPE = "area-shape"
    VALIDITY = "validity"
    SNAP_THRESHOLD = "snap-threshold"

    @classmethod
    def column_type(
        cls,
    ) -> Dict[ColumnNames, Union[Type[str], Type[float], Type[bool]]]:
        """
        Get python type for each column.
        """
        return {
            cls.AREA: str,
            cls.TRACES: str,
            cls.THEMATIC: str,
            cls.SCALE: str,
            cls.AREA_SHAPE: str,
            cls.VALIDITY: str,
            cls.SNAP_THRESHOLD: float,
        }


@unique
class AreaShapes(Enum):

    """
    area-shape options in database.csv.
    """

    CIRCLE = "circle"
    OTHER = "other"


@unique
class PathNames(Enum):

    """
    Path names.
    """

    DATA = "tracerepository_data"
    TRACES = "traces"
    AREA = "area"
    UNORGANIZED = "unorganized"
    REPORTS = "tracerepository_reports"
    METADATA = "metadata_rules.json"
    DATABASE_CSV = "database.csv"


@unique
class Geometry(Enum):

    """
    Folder names.
    """

    TRACES = ColumnNames.TRACES.value
    AREAS = ColumnNames.AREA.value
    BOTH = "both"


@unique
class ValidationResults(Enum):

    """
    Validation result description enums.
    """

    EMPTY = "empty"
    VALID = "valid"
    INVALID = "invalid"
    CRITICAL = "critical"
    UNFIT = "unfit"

    @classmethod
    def color_dict(cls) -> Dict[str, str]:
        """
        Assign colors to each enum value for cli reporting.
        """
        return {
            cls.EMPTY.value: colors.WHITE,
            cls.VALID.value: colors.GREEN,
            cls.INVALID.value: colors.RED,
            cls.CRITICAL.value: colors.BRIGHT_RED,
            cls.UNFIT.value: colors.YELLOW,
        }


def name_column_kwargs(
    geom_type: Optional[ColumnNames],
    allow_duplicates=True,
) -> Dict[str, Any]:
    """
    Get kwargs for a string/name column.
    """
    return dict(
        pandas_dtype=pa.String,
        checks=[
            pa.Check.str_matches(filename_regex(geom_type=geom_type)),
        ],
        allow_duplicates=allow_duplicates,
    )


def enum_column_kwargs(enum_class: Type[Enum]) -> Dict[str, Any]:
    """
    Get kwargs for an enum column.
    """
    return dict(
        pandas_dtype=pa.String,
        checks=pa.Check.isin([member.value for member in enum_class]),
    )


def filename_regex(geom_type: Optional[ColumnNames] = None) -> str:
    """
    Get general, trace or area filename regex.

    E.g.

    >>> filename_regex()
    '^[a-z0-9_]{2,49}$'

    >>> filename_regex(ColumnNames.AREA)
    '^[a-z0-9_]{2,49}_area$'

    >>> filename_regex(ColumnNames.TRACES)
    '^[a-z0-9_]{2,49}_traces$'

    """
    base = r"^[a-z0-9_]{2,49}"

    if geom_type is None:
        return base + r"$"
    if geom_type == ColumnNames.AREA:
        return base + r"_area$"
    if geom_type == ColumnNames.TRACES:
        return base + r"_traces$"
    raise TypeError(
        f"Expected geom_type {geom_type} to be None or TRACES or AREA enum."
    )


def database_schema() -> pa.DataFrameSchema:
    """
    Get pandera DataFrame schema for database.csv.
    """
    schema = pa.DataFrameSchema(
        # Index is the area name
        index=pa.Index(
            **name_column_kwargs(allow_duplicates=False, geom_type=ColumnNames.AREA)
        ),
        coerce=True,
        # Columns
        columns={
            # traces, thematic and scale columns are strings
            ColumnNames.TRACES.value: pa.Column(
                **name_column_kwargs(
                    allow_duplicates=True, geom_type=ColumnNames.TRACES
                )
            ),
            ColumnNames.THEMATIC.value: pa.Column(
                **name_column_kwargs(allow_duplicates=True, geom_type=None)
            ),
            ColumnNames.SCALE.value: pa.Column(
                **name_column_kwargs(allow_duplicates=True, geom_type=None)
            ),
            # area-shape must be one of the enum values
            ColumnNames.AREA_SHAPE.value: pa.Column(
                **enum_column_kwargs(enum_class=AreaShapes)
            ),
            # validated must be one of the enum values
            ColumnNames.VALIDITY.value: pa.Column(
                **enum_column_kwargs(enum_class=ValidationResults)
            ),
            ColumnNames.SNAP_THRESHOLD.value: pa.Column(
                pa.Float,
                checks=[
                    pa.Check.greater_than_or_equal_to(1e-8),
                    pa.Check.less_than_or_equal_to(1e8),
                ],
                coerce=True,
                nullable=False,
            ),
        },
    )
    assert isinstance(schema, pa.DataFrameSchema)
    return schema


# def folder_structure() -> List[Path]:
#     """
#     Get the default data folder structure.
#     """
#     return [Path(PathNames.DATA.value)]


class OrderedMeta(BaseModel):

    """
    Metadata construct for e.g. data_source and scale.
    """

    order: Dict[str, int]
    separator: str


class Metadata(BaseModel):

    """
    Metadata schema from json file for trace metadata validation.
    """

    certainty: Tuple[str, ...]
    operators: Tuple[str, ...]
    lineament_id_prefixes: Tuple[str, ...]
    data_source: OrderedMeta
    scale: OrderedMeta
    filepath: Path

    def __hash__(self):
        """
        Implement hashing using just the filepath.
        """
        return hash(self.filepath)



"""
Spatial data validation.
"""

import logging
from concurrent.futures import ProcessPoolExecutor, as_completed
from itertools import groupby
from pathlib import Path
from typing import List, Sequence, Tuple, Type, Union

import geopandas as gpd
from fractopo.general import is_empty_area, read_geofile
from fractopo.tval.trace_validation import Validation
from fractopo.tval.trace_validators import (
    ALL_VALIDATORS,
    BaseValidator,
    EmptyTargetAreaValidator,
    SharpCornerValidator,
)

from tracerepo import rules, utils
from tracerepo.rules import ValidationResults
from tracerepo.utils import TraceTuple

ANY_VALIDATOR = Union[Type[BaseValidator], Type[EmptyTargetAreaValidator]]

DRIVER_EXTENSIONS = {"ESRI Shapefile": ".shp", "GPKG": ".gpkg"}


def check_for_validator_error(
    errors: Tuple[str, ...], validators: Tuple[ANY_VALIDATOR, ...] = ALL_VALIDATORS
) -> bool:
    """
    Check if validator errors are in list of validation errors.
    """
    return any(validator.ERROR in errors for validator in validators)


def validate(
    traces: gpd.GeoDataFrame, area: gpd.GeoDataFrame, name: str, snap_threshold: float
) -> Tuple[gpd.GeoDataFrame, ValidationResults]:
    """
    Validate trace GeoDataFrame.
    """
    # Check for empty target area
    if is_empty_area(area=area, traces=traces):
        return traces.copy(), ValidationResults.EMPTY

    # Create Validation instance
    validation = Validation(
        name=name,
        traces=traces,
        area=area,
        allow_fix=True,
        SNAP_THRESHOLD=snap_threshold,
    )

    try:
        # Run validation
        validated = validation.run_validation(allow_empty_area=False)

    except Exception as exc:
        logging.critical(
            f"Validation critically failed for dataset ({name}) with exception: {exc}",
            exc_info=True,
        )
        return traces, ValidationResults.CRITICAL

    # Get the error column values
    validated_error_column_values = validated[validation.ERROR_COLUMN].values

    # Convert to list and make sure the error values are tuples (as of fractopo v0.3.0)
    validated_error_lists: List[Tuple[str, ...]] = [
        errors for errors in validated_error_column_values if isinstance(errors, tuple)
    ]

    # Make sure all were tuples
    assert len(validated_error_column_values) == len(validated_error_lists)

    # Check for major errors that require user fix
    if any(
        check_for_validator_error(
            errors,
            validators=tuple(
                validator
                for validator in ALL_VALIDATORS
                if validator is not SharpCornerValidator
            ),
        )
        for errors in validated_error_lists
    ):
        return validated, ValidationResults.INVALID
    # Traces are valid
    return validated, ValidationResults.VALID


def unique_invalids(invalids: List[utils.TraceTuple]) -> List[utils.TraceTuple]:
    """
    Return invalids that are unique by traces_path.
    """

    def keyfunc(invalid: utils.TraceTuple) -> Path:
        """
        Return traces_path from invalid.
        """
        traces_path = invalid.traces_path
        assert isinstance(traces_path, Path)
        return traces_path

    unique = []
    sorted_invalids = sorted(invalids, key=keyfunc)
    for _, group in groupby(sorted_invalids, key=keyfunc):
        unique.append(next(group))
    return unique


def sort_update_tuples_to_match_invalids(
    update_tuples: List[utils.UpdateTuple], invalids: Sequence[utils.TraceTuple]
) -> List[utils.UpdateTuple]:
    """
    Sort update_tuples to match the order in invalids based on area_name.
    """
    # Resolve the matching area name from area_path variable
    invalids_area_names = [invalid.area_path.stem for invalid in invalids]

    # Get the matching indexes for update_tuples
    idxs = [
        invalids_area_names.index(update_tuple.area_name)
        for update_tuple in update_tuples
    ]

    # Sort update_tuples with the idxs
    sorted_with_idx = sorted(zip(update_tuples, idxs), key=lambda vals: vals[1])

    # Extract only the update_tuples that should now be ordered
    sorted_update_tuples = [ut[0] for ut in sorted_with_idx]

    return sorted_update_tuples


def validate_invalids(invalids: Sequence[utils.TraceTuple]) -> List[utils.UpdateTuple]:
    """
    Validate a sequence of invalids with multiprocessing support.

    Will not validate the same trace dataset twice.
    """
    # Collect validated
    update_tuples: List[utils.UpdateTuple] = []

    # multiprocessing!
    with ProcessPoolExecutor() as executor:
        # Iterate over invalids. submit as tasks
        futures = {
            executor.submit(validate_invalid, invalid): invalid for invalid in invalids
        }

        # Collect all tasks as they complete
        # Will not be in same order submitted!?
        for future in as_completed(futures):

            # If validation critically fails for a dataset
            # we can still proceed with other validations
            try:

                # Get result from Future
                # This will throw an error (if it happened in process)
                update_tuple = future.result()
            except Exception:

                # Catch and log critical failures
                logging.error(
                    f"Validation exception with {futures[future]}.",
                    exc_info=True,
                )
                update_values = {
                    rules.ColumnNames.VALIDITY: rules.ValidationResults.CRITICAL.value
                }
                update_tuple = utils.UpdateTuple(
                    area_name=futures[future].area_path.stem,
                    update_values=update_values,
                    error=True,
                    traces_path=futures[future].traces_path,
                )
            # Collect result
            update_tuples.append(update_tuple)
    return sort_update_tuples_to_match_invalids(
        update_tuples=update_tuples, invalids=invalids
    )


def validate_invalid(invalid: utils.TraceTuple) -> utils.UpdateTuple:
    """
    Validate a given trace dataset.
    """
    # invalid is a TraceTuple which has named path attributes
    traces_path = invalid.traces_path
    area_path = invalid.area_path

    # Both should be Paths
    assert isinstance(traces_path, Path)
    assert isinstance(area_path, Path)

    # Read traces GeoDataFrame
    traces = read_geofile(traces_path)

    if traces.empty:
        return utils.UpdateTuple(
            area_name=area_path.stem,
            update_values={
                rules.ColumnNames.VALIDITY: rules.ValidationResults.EMPTY.value
            },
            traces_path=traces_path,
        )

    # Validate with fractopo trace validation
    validated, validation_results = validate(
        traces=traces,
        area=read_geofile(area_path),
        snap_threshold=invalid.snap_threshold,
        name=area_path.name,
    )

    if not validated.crs == traces.crs:
        raise ValueError("Expected crs to match for gdf before and after validation.")

    try:
        # Write the validated traces
        utils.write_geodata(gdf=validated, path=traces_path)
    except Exception:
        # Log exception
        logging.error(
            f"Error when writing validated trace GeoDataFrame to {traces_path}.",
            exc_info=True,
        )

        # Return with critical error
        # Validation does minor edits that are required for the dataset to be
        # valid so writing is always required.
        return utils.UpdateTuple(
            area_name=area_path.stem,
            update_values={
                rules.ColumnNames.VALIDITY: ValidationResults.CRITICAL.value
            },
            traces_path=traces_path,
        )

    # Create dict with information on validity for trace-area-combo
    update_tuple = utils.UpdateTuple(
        area_name=area_path.stem,
        update_values={rules.ColumnNames.VALIDITY: validation_results.value},
        traces_path=traces_path,
    )

    return update_tuple


def rename_trace_tuple_paths(
    trace_tuple: TraceTuple, export_destination: str, driver: str
) -> TraceTuple:
    """
    Rename TraceTuple paths with new base data directory and extension.
    """
    return TraceTuple(
        traces_path=rename_path(trace_tuple.traces_path, export_destination, driver),
        area_path=rename_path(trace_tuple.area_path, export_destination, driver),
        snap_threshold=trace_tuple.snap_threshold,
        validity=trace_tuple.validity,
    )


def rename_path(path: Path, export_destination: str, driver: str) -> Path:
    """
    Rename path to new base data dir and extension.

    >>> str(
    ...     rename_path(
    ...         Path("data/loviisa/traces/20m/hey.geojson"), "newdata", "ESRI Shapefile"
    ...     )
    ... )
     'newdata/loviisa/traces/20m/hey.shp'
    """
    renamed = utils.rename_data_path(path=path, rename_to=export_destination)

    # Convert suffix
    fully_renamed = renamed.with_suffix(DRIVER_EXTENSIONS[driver])

    assert DRIVER_EXTENSIONS[driver] in str(fully_renamed)
    assert export_destination in str(fully_renamed)

    return fully_renamed


def convert_trace_tuples(
    trace_tuples: Sequence[TraceTuple], export_destination: str, driver: str
) -> List[TraceTuple]:
    """
    Make paths for converting between geodata filetypes.

    >>> from pprint import pprint
    >>> trace_tuple = TraceTuple(
    ...     traces_path=Path("data/loviisa/traces/20m/traces.geojson"),
    ...     area_path=Path("data/loviisa/traces/20m/area.geojson"),
    ...     validity=rules.ValidationResults.VALID.value,
    ... )
    >>> converted = convert_trace_tuples([trace_tuple], "exported", "GPKG")[0]
    >>> pprint(list(converted.traces_path.parents))
    [PosixPath('exported/loviisa/traces/20m'),
     PosixPath('exported/loviisa/traces'),
     PosixPath('exported/loviisa'),
     PosixPath('exported'),
     PosixPath('.')]
    """
    convert_paths: List[TraceTuple] = []
    # Iterate over datasets and save
    trace_tuple: TraceTuple
    for trace_tuple in trace_tuples:
        convert_paths.append(
            rename_trace_tuple_paths(
                trace_tuple=trace_tuple,
                export_destination=export_destination,
                driver=driver,
            )
        )

    return convert_paths


def save_converted_paths(
    src_trace_tuples: Sequence[TraceTuple],
    dest_trace_tuples: Sequence[TraceTuple],
    driver: str,
    destination: Path,
):
    """
    Save transformed geodata files to new paths.
    """
    for src_trace_tuple, dest_trace_tuple in zip(src_trace_tuples, dest_trace_tuples):

        for original_path, convert_path in zip(
            (src_trace_tuple.traces_path, src_trace_tuple.area_path),
            (dest_trace_tuple.traces_path, dest_trace_tuple.area_path),
        ):
            convert_filetype(original_path, destination / convert_path, driver=driver)


def convert_filetype(original_path: Path, convert_path: Path, driver: str):
    """
    Convert from original_path to convert_path with driver.
    """
    if not original_path.exists():
        raise FileNotFoundError(
            f"Expected {original_path.name} to exist at {original_path}."
        )

    # If already saved same trace dataset no need to overwrite
    # If exporting to directory that you've already previously exported to
    # it is removed before exporting
    if convert_path.exists():
        logging.info(f"Dataset already exists at {convert_path}.")
        return

    # Read from path
    gdf = read_geofile(original_path)

    # Make parent directories as needed
    convert_path.parent.mkdir(exist_ok=True, parents=True)

    # Save with new extension and type
    logging.info(f"Saving to {convert_path} with driver {driver}.")
    try:
        gdf.to_file(convert_path, driver=driver)
    except Exception:
        logging.error(
            f"Failed to save {original_path} to {convert_path}"
            f" with driver {driver} due to error.",
            exc_info=True,
        )



"""
Trace schema checks used in validation.
"""
import re
from typing import Any, Dict, Tuple

import numpy as np

LINEAMENT_ID_PATTERN_UNFORMATTED = r"^({})_\w\d+$"


def pattern_matcher(value: str, pattern_str: str) -> bool:
    """
    Check if string matches regex pattern.
    """
    pattern = re.compile(pattern_str)
    try:
        pattern_match = pattern.match(value)
    except TypeError:
        return False
    return pattern_match is not None


def named_priority_check(value: Any, named_priorities: Dict[str, int], separator: str):
    """
    Check that value matches a predetermined name and priority order of names.
    """
    if len(value) == 0 or not isinstance(value, str):
        return False
    # Check if value matches one of the named
    if value in named_priorities:
        return True
    # Check if any of the named are actually in the value string
    if (not any(name in value for name in named_priorities)) or separator not in value:
        return False
    # Check if separator is within the value string
    # if separator not in value:
    #     return False
    # Split by separator
    split_value = value.split(sep=separator)

    # if not all(val in named_priorities for val in split_value):
    #     return False

    previous = 0
    for part in split_value:
        # Check that all in split value are in named_priorities
        if part not in named_priorities:
            return False
        current = named_priorities[part]
        if current > previous:
            previous = current
        else:
            return False
    return True


def date_datetime_check(raw_value: Any) -> bool:
    """
    Check that date column has valid datetime values.

    >>> date_datetime_check(np.datetime64("2021-07-23"))
    True

    >>> date_datetime_check(np.datetime64("1995-07-23"))
    False

    """
    try:
        if not isinstance(raw_value, np.datetime64):
            value = np.datetime64(raw_value)
        else:
            value = raw_value
        after_2000 = value > np.datetime64("2000-01-01")
        before_2100 = value < np.datetime64("2100-01-01")

        return all((after_2000, before_2100))
    except (ValueError, TypeError):
        return False


def lineament_id_check(raw_value: Any, lineament_id_prefixes: Tuple[str, ...]) -> bool:
    """
    Check that Lineament_ID column value matches predefined pattern.

    >>> lineament_id_check("M_A1", ("M",))
    True

    >>> lineament_id_check("M_A1", ("M", "L", "EM"))
    True

    >>> lineament_id_check("L_A1", ("M", "L", "EM"))
    True

    >>> lineament_id_check("H_A1", ("M", "L", "EM"))
    False
    """
    if not isinstance(raw_value, str) or len(raw_value) == 0:
        return False
    compiled_pattern = re.compile(
        LINEAMENT_ID_PATTERN_UNFORMATTED.format("|".join(lineament_id_prefixes))
    )
    return compiled_pattern.match(raw_value) is not None



"""
Command line api for tracerepo.
"""

import logging
from pathlib import Path
from shutil import rmtree
from typing import List, Optional, Sequence

import typer
from json5 import loads
from nialog.logger import setup_module_logging
from rich.console import Console
from rich.text import Text

from tracerepo import repo, rules, spatial, utils
from tracerepo.organize import Organizer

app = typer.Typer()
console = Console()

DATABASE_OPTION = typer.Option(
    rules.DATABASE_CSV,
)

TRACEREPOSITORY_PATH_OPTION = typer.Option(
    ".",
    exists=True,
    file_okay=False,
    dir_okay=True,
)

DATA_FILTER = typer.Option(default=())


def logging_level(level: str):
    """
    Make logging level string.
    """
    return f"Set logging level to {level}."


@app.callback()
def app_callback(
    verbose: bool = typer.Option(False, help=logging_level("INFO")),
    debug: bool = typer.Option(False, help=logging_level("DEBUG")),
):
    """
    Use tracerepo to manage and validate fracture & lineament trace data.
    """
    logging_level_int = logging.WARNING
    # logging.basicConfig(level=logging.WARNING, force=True)
    if verbose:
        # logging.basicConfig(level=logging.INFO, force=True)
        logging_level_int = logging.INFO
    if debug:
        # logging.basicConfig(level=logging.DEBUG, force=True)
        logging_level_int = logging.DEBUG
    setup_module_logging(logging_level_int=logging_level_int)
    logging.info(
        "Logging verbosity set and nialog initialized.",
        extra=dict(logging_level_int=logging_level_int, verbose=verbose, debug=debug),
    )


def load_metadata_from_json(metadata_json_path: Path) -> rules.Metadata:
    """
    Load and parse json metadata of trace columns.
    """
    # Load metadata of column restrictions
    loaded_metadata = loads(metadata_json_path.read_text())
    if not isinstance(loaded_metadata, dict):
        raise TypeError(
            f"Expected {metadata_json_path} to parse as a dict."
            f" Got {type(loaded_metadata)}."
        )
    return rules.Metadata(**loaded_metadata, filepath=metadata_json_path)


@app.command()
def validate(
    tracerepository_path: Path = TRACEREPOSITORY_PATH_OPTION,
    database_name: str = DATABASE_OPTION,
    area_filter: List[str] = DATA_FILTER,
    thematic_filter: List[str] = DATA_FILTER,
    traces_filter: List[str] = DATA_FILTER,
    scale_filter: List[str] = DATA_FILTER,
    report: bool = typer.Option(False),
    # report_directory: Optional[Path] = typer.Option(rules.PathNames.REPORTS.value),
    report_directory: Optional[Path] = typer.Option(
        None,
        help=(
            "Defaults to file in tracerepository_path directory "
            f"with name: {rules.PathNames.REPORTS.value}."
        ),
    ),
    metadata_json: Optional[Path] = typer.Option(
        None,
        help=(
            "Defaults to file in tracerepository_path directory "
            f"with name: {rules.PathNames.METADATA.value}."
        ),
        # rules.PathNames.METADATA.value, exists=True, dir_okay=False
    ),
):
    """
    Validate trace datasets.

    Only validates if the dataset has been marked as invalid in database.csv.
    """
    console = Console()
    database = tracerepository_path / database_name
    # Initialize Organizer
    organizer = Organizer(
        tracerepository_path=tracerepository_path,
        database=repo.read_database_csv(path=database),
    )

    # Resolve metadata_json_path
    metadata_json_path = (
        metadata_json
        if metadata_json is not None
        else tracerepository_path / rules.PathNames.METADATA.value
    )

    # Load metadata of traces column restrictions
    metadata = load_metadata_from_json(metadata_json_path=metadata_json_path)

    # Query for invalid traces
    invalids = organizer.query(
        area=area_filter,
        thematic=thematic_filter,
        traces=traces_filter,
        scale=scale_filter,
        validity=[
            rules.ValidationResults.INVALID,
            rules.ValidationResults.EMPTY,
            rules.ValidationResults.UNFIT,
        ],
    )

    # Only validate a single trace dataset once
    # Means you might have to validate for each area dataset
    # that uses the traces.
    unique_invalids_only = spatial.unique_invalids(invalids=invalids)

    # Report which data are validated
    if report:
        console.print(utils.create_validation_table(unique_invalids_only))

    # Validate the invalids
    update_tuples = spatial.validate_invalids(invalids=unique_invalids_only)

    # Exit with error code 1 if there's errors in updating the database.csv
    database_error, write_error = False, False

    assert len(update_tuples) == len(unique_invalids_only)
    # Iterate over results

    for update_tuple, invalid in zip(update_tuples, unique_invalids_only):

        # Validate and gather pandera reporting
        pandera_update_values, pandera_report = utils.pandera_reporting(
            update_tuple=update_tuple,
            metadata=metadata,
        )

        # If the geodataset is otherwise valid but fails pandera checks it will
        # be marked as unfit
        if len(pandera_update_values) > 0:
            update_tuple.update_values = pandera_update_values
        try:
            # Update Organizer database.csv
            organizer.update(
                area_name=invalid.area_path.stem,
                update_values=update_tuple.update_values,
            )

            # Write the database.csv
            repo.write_database_csv(path=database, database=organizer.database)

        except Exception:

            # Error in updating or writing database.csv
            database_error = True

            # Log exception
            logging.error(
                f"Error when updating or writing Organizer database.csv.\n"
                f"update_tuple: {update_tuple}\n"
                f"invalid: {invalid}\n",
                exc_info=True,
            )

        if not pandera_report.empty and report:
            report_directory = (
                tracerepository_path / Path(rules.PathNames.REPORTS.value)
                if report_directory is None
                else report_directory
            )
            str_report = utils.report_pandera_errors(
                pandera_report=pandera_report,
                report_directory=report_directory,
                area_name=update_tuple.area_name,
            )
            console.print(str_report)

    # Report validation results with a rich.table.Table
    if report:
        console.print(
            utils.create_validation_results_table(
                invalids=unique_invalids_only, update_tuples=update_tuples
            )
        )

    if database_error or write_error:

        # Exit with error code 1 (not successful)
        raise typer.Exit(code=1)


@app.command()
def format_geojson(
    tracerepository_path: Path = TRACEREPOSITORY_PATH_OPTION,
    database_name: str = DATABASE_OPTION,
):
    """
    Format all dataset GeoJSON from cli.
    """
    database = tracerepository_path / database_name
    format_repo_geojson(database=database, tracerepository_path=tracerepository_path)


def format_repo_geojson(tracerepository_path: Path, database: Path):
    """
    Format all dataset GeoJSON.
    """
    # Initialize Organizer
    organizer = Organizer(
        tracerepository_path=tracerepository_path,
        database=repo.read_database_csv(path=database),
    )

    # Query for invalid traces
    trace_tuples = organizer.query()

    # Do not load and save same paths multiple times
    formatted_paths = set()

    # Iterate over all area trace tuples

    for trace_tuple in trace_tuples:

        for path in (trace_tuple.traces_path, trace_tuple.area_path):
            if path in formatted_paths:
                continue

            # Read GeoDataFrame
            gdf = spatial.read_geofile(path)

            # Write as GeoJSON
            utils.write_geodata(gdf=gdf, path=path)

            # Add formatted paths
            formatted_paths.add(path)


@app.command()
def organize(
    tracerepository_path: Path = TRACEREPOSITORY_PATH_OPTION,
    database_name: str = DATABASE_OPTION,
    simulate: bool = typer.Option(False),
    report: bool = typer.Option(True),
):
    """
    Organize repo.
    """
    database = tracerepository_path / database_name
    organizer = Organizer(
        tracerepository_path=tracerepository_path,
        database=repo.read_database_csv(path=database),
    )

    move_descriptions = organizer.organize(simulate=simulate)

    if report:
        console.print(Text("\n".join(move_descriptions), style="yellow"))

    if not simulate:
        organizer.check()


@app.command()
def check(
    tracerepository_path: Path = TRACEREPOSITORY_PATH_OPTION,
    database_name: str = DATABASE_OPTION,
):
    """
    Check repo.
    """
    database = tracerepository_path / database_name
    organizer = Organizer(
        tracerepository_path=tracerepository_path,
        database=repo.read_database_csv(path=database),
    )

    organizer.check()


@app.command()
def init(
    tracerepository_path: Path = TRACEREPOSITORY_PATH_OPTION,
    database_name: str = DATABASE_OPTION,
):
    """
    Initialize tracerepo in current directory.
    """
    df = repo.scaffold(tracerepository_path=tracerepository_path)
    repo.write_database_csv(path=tracerepository_path / database_name, database=df)


def export_data(
    destination: Path,
    driver: str,
    database: Path,
    tracerepository_path: Path,
    area_filter: Sequence[str] = (),
    thematic_filter: Sequence[str] = (),
    traces_filter: Sequence[str] = (),
    scale_filter: Sequence[str] = (),
    overwrite: bool = True,
) -> Path:
    """
    Export datasets into another format.
    """
    assert destination.is_dir()
    # Initialize Organizer
    organizer = Organizer(
        tracerepository_path=tracerepository_path,
        database=repo.read_database_csv(path=database),
    )

    # Query for datasets based on filters
    # By default filters are empty i.e. all are selected
    trace_tuples = organizer.query(
        area=area_filter,
        thematic=thematic_filter,
        traces=traces_filter,
        scale=scale_filter,
    )

    # Resolve the export destination folder
    export_destination_dir = utils.compile_export_dir(driver)
    export_destination_path = destination / export_destination_dir

    # Does destination already exist
    destination_exists = export_destination_path.exists()

    if destination_exists and overwrite:

        logging.info(f"Removing directory ({export_destination_path}) recursively.")
        rmtree(export_destination_path)

    elif destination_exists:
        raise FileExistsError(
            f"Directory already exists at {export_destination_path} "
            "and overwrite is not allowed (--no-overwrite given)."
        )

    # Compile from trace tuples into paths
    dest_trace_tuples = spatial.convert_trace_tuples(
        trace_tuples, export_destination=export_destination_dir, driver=driver
    )

    # Export to disk
    spatial.save_converted_paths(
        src_trace_tuples=trace_tuples,
        dest_trace_tuples=dest_trace_tuples,
        driver=driver,
        destination=destination,
    )

    return export_destination_path


@app.command()
def export(
    destination: Path = typer.Argument(".", file_okay=False),
    driver: str = typer.Option("ESRI Shapefile"),
    tracerepository_path: Path = TRACEREPOSITORY_PATH_OPTION,
    database_name: str = DATABASE_OPTION,
    area_filter: List[str] = DATA_FILTER,
    thematic_filter: List[str] = DATA_FILTER,
    traces_filter: List[str] = DATA_FILTER,
    scale_filter: List[str] = DATA_FILTER,
    overwrite: bool = typer.Option(True),
):
    """
    Export datasets into another format from command line.
    """
    database = tracerepository_path / database_name
    export_destination = export_data(
        tracerepository_path=tracerepository_path,
        destination=destination,
        driver=driver,
        database=database,
        area_filter=area_filter,
        thematic_filter=thematic_filter,
        traces_filter=traces_filter,
        scale_filter=scale_filter,
        overwrite=overwrite,
    )

    text = Text(
        f"Saved datasets to {export_destination} with driver {driver}.", style="green"
    )

    console.print(text)



"""
Main repo handlers.
"""
from pathlib import Path

import pandas as pd

from tracerepo import rules


def write_database_csv(path: Path, database: pd.DataFrame):
    """
    Write database.csv to disk.
    """
    database = rules.database_schema().validate(database)
    database.to_csv(
        path_or_buf=path,
        sep=rules.DATABASE_CSV_SEP,
        index=True,
        index_label=rules.ColumnNames.AREA.value,
    )


def read_database_csv(path: Path) -> pd.DataFrame:
    """
    Read database csv.
    """
    csv = pd.read_csv(path, index_col=0, sep=rules.DATABASE_CSV_SEP, dtype=str)
    assert isinstance(csv, pd.DataFrame)
    csv = rules.database_schema().validate(csv)
    return csv


def scaffold_database():
    """
    Make scaffold database.
    """
    return pd.DataFrame(columns=[col.value for col in rules.ColumnNames]).set_index(
        rules.ColumnNames.AREA.value
    )


def scaffold(tracerepository_path: Path) -> pd.DataFrame:
    """
    Make scaffold start for a repo.
    """
    unorganized_path = tracerepository_path / rules.PathNames.UNORGANIZED.value
    data_path = tracerepository_path / rules.PathNames.DATA.value
    # Make unorganized folder
    unorganized_path.mkdir(exist_ok=True)
    data_path.mkdir(exist_ok=True)

    # # Make other default folders

    # for path in rules.folder_structure():
    #     path.mkdir(exist_ok=True, parents=True)

    # Create dataframe for relations between datasets
    df = scaffold_database()
    assert df is not None

    # Validate dataframe with pandera
    df = rules.database_schema().validate(df)
    return df



"""
Organize trace data.
"""

from dataclasses import dataclass
from itertools import compress
from pathlib import Path
from shutil import move
from typing import Any, Dict, List, Optional, Sequence

import pandas as pd

from tracerepo import rules, utils


@dataclass
class Organizer:

    """
    Organize trace data files with Organizer.
    """

    database: pd.DataFrame
    tracerepository_path: Path

    _columns: Optional[Dict[str, List[Any]]] = None

    def __post_init__(self):
        """
        Post initialization steps.
        """
        assert self.tracerepository_path.exists()
        self.database = rules.database_schema().validate(self.database)
        self.unorganized_folder = (
            self.tracerepository_path / rules.PathNames.UNORGANIZED.value
        )

    @property
    def unorganized(self) -> List[Path]:
        """
        Find unorganized files in unorganized_folder.
        """
        return list(self.unorganized_folder.glob(f"*.{rules.FILETYPE}"))

    def organize(self, simulate: bool = False) -> List[str]:
        """
        Organize files from unorganized_folder.
        """
        move_descriptions = []

        for filepath in self.unorganized:
            filename_stem = filepath.stem
            geometry = utils.identify_geom_type(filename_stem).value
            idx_in_database = self.columns[geometry].index(filename_stem)
            thematic = self.columns[rules.ColumnNames.THEMATIC.value][idx_in_database]
            scale = self.columns[rules.ColumnNames.SCALE.value][idx_in_database]
            destination = utils.compiled_path(
                root=self.tracerepository_path,
                thematic=thematic,
                geometry=geometry,
                scale=scale,
                name=filename_stem,
            )
            if not simulate:
                destination.parent.mkdir(parents=True, exist_ok=True)
                move(str(filepath.absolute().expanduser()), destination)
            move_description = f"Moving {filepath} to {destination}." + (
                " --SIMULATION--" if simulate else ""
            )
            assert isinstance(move_description, str)
            move_descriptions.append(move_description)
        return move_descriptions

    @property
    def columns(self) -> Dict[str, List[Any]]:
        """
        Get database columns and index as Python typed values.

        Accessible as a dict with column names as keys. Includes the area name
        values that were originally the dataframe index.
        """
        if self._columns is None:
            cols: Dict[str, List[Any]] = dict()
            for column in rules.ColumnNames:
                cols[column.value] = utils.dataframe_column_to_python(
                    dataframe=self.database,
                    column=column.value,
                    python_type=rules.ColumnNames.column_type()[column],
                )
            self._columns = cols
        return self._columns

    def check(self):
        """
        Check if all rows in database correspond to files.

        Furthermore check if all files in data dir correspond to rows.
        """
        data_path = self.tracerepository_path / Path(rules.PathNames.DATA.value)
        all_files_and_dirs = list(data_path.rglob("*"))
        all_files = {path.stem: path for path in all_files_and_dirs if path.is_file()}
        all_dirs = {path.stem: path for path in all_files_and_dirs if path.is_dir()}

        for value in (rules.PathNames.AREA.value, rules.PathNames.TRACES.value):
            all_dirs = utils.remove_from_dict_if_in(key=value, dict_to_check=all_dirs)

        for area, traces, thematic, scale in zip(
            self.columns[rules.ColumnNames.AREA.value],
            self.columns[rules.ColumnNames.TRACES.value],
            self.columns[rules.ColumnNames.THEMATIC.value],
            self.columns[rules.ColumnNames.SCALE.value],
        ):

            for value in (thematic, scale):
                all_dirs = utils.remove_from_dict_if_in(
                    key=value, dict_to_check=all_dirs
                )
            for value in (area, traces):
                all_files = utils.remove_from_dict_if_in(
                    key=value, dict_to_check=all_files
                )

            for geom_filename, geometry in zip(
                (area, traces),
                (rules.ColumnNames.AREA.value, rules.ColumnNames.TRACES.value),
            ):
                utils.check_database_row_files(
                    root=self.tracerepository_path,
                    thematic=thematic,
                    geometry=geometry,
                    scale=scale,
                    name=geom_filename,
                )
        orphan_files = len(all_files) != 0
        orphan_dirs = len(all_dirs) != 0

        if orphan_files or orphan_dirs:
            orphan_file_error_str = f"Found orphan files: {list( all_files.values() )}."
            orphan_dir_error_str = (
                f"Found orphan directories: {list( all_dirs.values() )}."
            )

            raise FileExistsError(
                f"Expected all files and directories in "
                f"{data_path.absolute()}"
                " to correspond to row values in database.\n"
                f"{orphan_file_error_str}\n"
                f"{orphan_dir_error_str}\n"
            )

    @staticmethod
    def _filter_strings(
        area_values: Sequence[str],
        traces_values: Sequence[str],
        thematic_values: Sequence[str],
        scale_values: Sequence[str],
        query_bools: Sequence[bool],
        area: Sequence[str] = (),
        traces: Sequence[str] = (),
        thematic: Sequence[str] = (),
        scale: Sequence[str] = (),
    ) -> Sequence[bool]:
        """
        Filter database traces and areas based on given strings.
        """
        for filterer, list_to_filter in zip(
            (area, traces, thematic, scale),
            (area_values, traces_values, thematic_values, scale_values),
        ):
            filtered = utils.multi_string_filter(
                strings=filterer, list_to_filter=list_to_filter
            )

            query_bools = utils.join_bools(filtered, query_bools)
        return query_bools

    @staticmethod
    def _filter_enums(
        area_shape_values: Sequence[str],
        validity_values: Sequence[str],
        query_bools: Sequence[bool],
        area_shape: Sequence[rules.AreaShapes] = (),
        validity: Sequence[rules.ValidationResults] = (),
    ) -> Sequence[bool]:
        """
        Filter database traces and areas based on given enum choices.
        """
        for filterer, list_to_filter in zip(
            (area_shape, validity),
            (
                area_shape_values,
                validity_values,
            ),
        ):
            assert isinstance(filterer, Sequence)
            if filterer is None or len(filterer) == 0:
                continue

            assert isinstance(filterer, Sequence)
            assert None not in filterer
            # filterer_vals = [filterer_val.value for filterer_val in filterer]
            # boolean = lambda val: val in filterer_vals or len(filterer_vals) == 0
            # else:
            # boolean = lambda val: val == filterer.value

            filterer_value_set = set(filt.value for filt in filterer)

            query_bools = [
                all(
                    (
                        val in filterer_value_set,
                        query_bool_val,
                    )
                )
                for val, query_bool_val in zip(list_to_filter, query_bools)
            ]
        return query_bools

    def query(
        self,
        area: Sequence[str] = (),
        traces: Sequence[str] = (),
        thematic: Sequence[str] = (),
        scale: Sequence[str] = (),
        area_shape: Sequence[rules.AreaShapes] = (),
        validity: Sequence[rules.ValidationResults] = (),
    ) -> List[utils.TraceTuple]:
        """
        Query for trace and area data.
        """
        # default value, all accepted
        query_bools = [True] * len(self.columns[rules.ColumnNames.AREA.value])

        # Check area, traces, thematic and scale filters
        query_bools = self._filter_strings(
            area_values=self.columns[rules.ColumnNames.AREA.value],
            traces_values=self.columns[rules.ColumnNames.TRACES.value],
            thematic_values=self.columns[rules.ColumnNames.THEMATIC.value],
            scale_values=self.columns[rules.ColumnNames.SCALE.value],
            query_bools=query_bools,
            area=area,
            traces=traces,
            thematic=thematic,
            scale=scale,
        )

        # Return empty if none accepted already
        if not any(query_bools):
            return []

        # Check area_shape, empty and validated filters
        query_bools = self._filter_enums(
            query_bools=query_bools,
            area_shape_values=self.columns[rules.ColumnNames.AREA_SHAPE.value],
            validity_values=self.columns[rules.ColumnNames.VALIDITY.value],
            area_shape=area_shape,
            validity=validity,
        )

        # Return if no accepted
        if not any(query_bools):
            return []

        # Collect the columns that are needed for path solving
        columns_needed = [
            rules.ColumnNames.THEMATIC.value,
            rules.ColumnNames.SCALE.value,
            rules.ColumnNames.TRACES.value,
            rules.ColumnNames.AREA.value,
            rules.ColumnNames.SNAP_THRESHOLD.value,
            rules.ColumnNames.VALIDITY.value,
        ]

        # Compress the values corresponding to columns_needed based on
        # query_bools boolean list
        (
            thematic_vals,
            scale_vals,
            trace_vals,
            area_vals,
            snap_vals,
            validity_vals,
        ) = tuple(
            list(compress(data=self.columns[col], selectors=query_bools))
            for col in columns_needed
        )

        # Collect trace and area paths (both or one of depending on geometry
        # filter) into named tuples.
        paths: List[utils.TraceTuple] = [
            utils.query_result_tuple(
                tracerepository_path=self.tracerepository_path,
                thematic_val=thematic_val,
                scale_val=scale_val,
                traces_val=traces_val,
                area_val=area_val,
                snap_threshold=snap_val,
                validity_val=validity_val,
            )
            for (
                thematic_val,
                scale_val,
                traces_val,
                area_val,
                snap_val,
                validity_val,
            ) in zip(
                thematic_vals,
                scale_vals,
                trace_vals,
                area_vals,
                snap_vals,
                validity_vals,
            )
        ]

        return paths

    def update(self, area_name: str, update_values: Dict[rules.ColumnNames, str]):
        """
        Change a value in the database.
        """
        # Get index for area_name
        index_for_area = self.columns[rules.ColumnNames.AREA.value].index(area_name)

        # Iterate over keys in update_values dict
        for key in update_values:

            # Get the current values
            column_values = self.columns[key.value]

            # Update value at index
            column_values[index_for_area] = update_values[key]

            # Make copy of instance dataframe
            database = self.database.copy()

            # Update column in dataframe with updated value
            database[key.value] = column_values

            # Validate
            database = rules.database_schema().validate(database)

            # Set new database
            self.database = database

            # Reset columns cached property
            self._columns = None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tracerepo']

package_data = \
{'': ['*']}

install_requires = \
['fractopo>=0.3.0,<0.4.0',
 'json5>=0.9.6,<0.10.0',
 'nialog>=0.0.1,<0.0.2',
 'pandas',
 'pandera',
 'pydantic>=1.8.2,<2.0.0',
 'pyproj>=3.1,<3.2',
 'rich',
 'typer']

extras_require = \
{'coverage': ['coverage>=5.0,<6.0', 'coverage-badge'],
 'docs': ['sphinx',
          'sphinx-rtd-theme',
          'nbsphinx',
          'sphinx-gallery',
          'sphinx-autodoc-typehints',
          'sphinx-autobuild'],
 'format-lint': ['sphinx',
                 'pylint',
                 'rstcheck',
                 'black[jupyter]',
                 'blacken-docs',
                 'blackdoc',
                 'isort'],
 'typecheck': ['mypy']}

entry_points = \
{'console_scripts': ['tracerepo = tracerepo.cli:app']}

setup_kwargs = {
    'name': 'tracerepo',
    'version': '0.0.7',
    'description': 'Fracture & lineament data management.',
    'long_description': 'Documentation\n=============\n\n|Documentation Status| |PyPI Status| |CI Test| |Coverage|\n\nRunning tests\n-------------\n\nTo run pytest in currently installed environment:\n\n.. code:: bash\n\n   poetry run pytest\n\nTo run full extensive test suite:\n\n.. code:: bash\n\n   poetry run doit\n   # Add -n <your-cpu-core-count> to execute tasks in parallel\n   # E.g.\n   poetry run doit -n 8 -v 0\n   # -v 0 is added to limit verbosity to mininum (optional)\n   # doit makes sure tasks are run in the correct order\n   # E.g. if a task uses a requirements.txt file that other task produces\n   # the producer is run first even with parallel execution\n\nFormatting and linting\n----------------------\n\n\nFormatting & linting:\n\n.. code:: bash\n\n   poetry run doit format\n   poetry run doit lint\n\nBuilding docs\n-------------\n\nDocs can be built locally to test that ``ReadTheDocs`` can also build them:\n\n.. code:: bash\n\n   poetry run doit docs\n\ndoit usage\n----------\n\nTo list all available commands from ``dodo.py``:\n\n.. code:: bash\n\n   poetry run doit list\n\nDevelopment\n~~~~~~~~~~~\n\nDevelopment dependencies for ``tracerepo`` include:\n\n-  ``poetry``\n\n   -  Used to handle Python package dependencies.\n\n   .. code:: bash\n\n      # Use poetry run to execute poetry installed cli tools such as invoke,\n      # nox and pytest.\n      poetry run <cmd>\n\n\n-  ``doit``\n\n   -  A general task executor that is a replacement for a ``Makefile``\n   -  Understands task dependencies and can run tasks in parallel\n      even while running them in the order determined from dependencies\n      between tasks. E.g. requirements.txt is a requirement for running\n      tests and therefore the task creating requirements.txt will always\n      run before the test task.\n\n   .. code:: bash\n\n      # Tasks are defined in dodo.py\n      # To list doit tasks from command line\n      poetry run doit list\n      # To run all tasks in parallel (recommended before pushing and/or\n      # committing)\n      # 8 is the number of cpu cores, change as wanted\n      # -v 0 sets verbosity to very low. (Errors will always still be printed.)\n      poetry run doit -n 8 -v 0\n\n-  ``nox``\n\n   -  ``nox`` is a replacement for ``tox``. Both are made to create\n      reproducible Python environments for testing, making docs locally, etc.\n\n   .. code:: bash\n\n      # To list available nox sessions\n      # Sessions are defined in noxfile.py\n      poetry run nox --list\n\n-  ``copier``\n\n   -  ``copier`` is a project templater. Many Python projects follow a similar\n      framework for testing, creating documentations and overall placement of\n      files and configuration. ``copier`` allows creating a template project\n      (e.g. https://github.com/nialov/nialov-py-template) which can be firstly\n      cloned as the framework for your own package and secondly to pull updates\n      from the template to your already started project.\n\n   .. code:: bash\n\n      # To pull copier update from github/nialov/nialov-py-template\n      poetry run copier update\n\n\n-  ``pytest``\n\n   -  ``pytest`` is a Python test runner. It is used to run defined tests to\n      check that the package executes as expected. The defined tests in\n      ``./tests`` contain many regression tests (done with\n      ``pytest-regressions``) that make it almost impossible\n      to add features to ``tracerepo`` that changes the results of functions\n      and methods.\n\n   .. code:: bash\n\n      # To run tests implemented in ./tests directory and as doctests\n      # within project itself:\n      poetry run pytest\n\n\n-  ``coverage``\n\n   .. code:: bash\n\n      # To check coverage of tests\n      # (Implemented as nox session!)\n      poetry run nox --session test_pip\n\n-  ``sphinx``\n\n   -  Creates documentation from files in ``./docs_src``.\n\n   .. code:: bash\n\n      # To create documentation\n      # (Implemented as nox session!)\n      poetry run nox --session docs\n\nBig thanks to all maintainers of the above packages!\n\nLicense\n~~~~~~~\n\nCopyright © 2021, Nikolas Ovaskainen.\n\n-----\n\n\n.. |Documentation Status| image:: https://readthedocs.org/projects/tracerepo/badge/?version=latest\n   :target: https://tracerepo.readthedocs.io/en/latest/?badge=latest\n.. |PyPI Status| image:: https://img.shields.io/pypi/v/tracerepo.svg\n   :target: https://pypi.python.org/pypi/tracerepo\n.. |CI Test| image:: https://github.com/nialov/tracerepo/workflows/test-and-publish/badge.svg\n   :target: https://github.com/nialov/tracerepo/actions/workflows/test-and-publish.yaml?query=branch%3Amaster\n.. |Coverage| image:: https://raw.githubusercontent.com/nialov/tracerepo/master/docs_src/imgs/coverage.svg\n   :target: https://github.com/nialov/tracerepo/blob/master/docs_src/imgs/coverage.svg\n',
    'author': 'nialov',
    'author_email': 'nikolasovaskainen@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nialov/tracerepo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



"""
Traces pandera scheming.
"""
from functools import lru_cache
from typing import Dict

import pandera as pa
from fractopo.tval.trace_validation import Validation

from tracerepo import rules, schema_checks

VALIDATION_ERROR_COLUMN = Validation.ERROR_COLUMN
DIP_COLUMN = "DIP"
DIP_DIR_COLUMN = "DIP_DIR"
DATA_SOURCE_COLUMN = "Data_Source"
DATE_COLUMN = "Date"
OPERATOR_COLUMN = "Operator"
SCALE_COLUMN = "Scale"
CERTAINTY_COLUMN = "Certainty"
LINEAMENT_ID_COLUMN = "Lineament_ID"


def default_non_required_kwargs(
    required: bool = False, coerce: bool = True, nullable: bool = True
) -> Dict[str, bool]:
    """
    Get default non required, coercible and nullable column kwargs.
    """
    return dict(required=required, coerce=coerce, nullable=nullable)


def prioritized_values_check(
    named_priorities: Dict[str, int], separator: str, name: str
) -> pa.checks.Check:
    """
    Construct check for e.g. data source and scale columns.

    Both have fixed values and fixed order of the values.
    """
    return pa.Check(
        lambda value: schema_checks.named_priority_check(
            value,
            named_priorities=named_priorities,
            separator=separator,
        ),
        element_wise=True,
        name=name,
    )


@lru_cache(maxsize=None)
def traces_schema(metadata: rules.Metadata):
    """
    Get pandera schema for traces GeoDataFrame.
    """
    trace_columns: Dict[str, pa.Column] = {
        VALIDATION_ERROR_COLUMN: pa.Column(pa.String, **default_non_required_kwargs()),
        DIP_COLUMN: pa.Column(
            pa.Float,
            **default_non_required_kwargs(),
            checks=[pa.checks.Check.in_range(min_value=0.0, max_value=90.0)],
        ),
        DIP_DIR_COLUMN: pa.Column(
            pa.Float,
            **default_non_required_kwargs(),
            checks=[pa.checks.Check.in_range(min_value=0.0, max_value=360.0)],
        ),
        DATA_SOURCE_COLUMN: pa.Column(
            pa.String,
            **default_non_required_kwargs(nullable=False),
            checks=[
                prioritized_values_check(
                    named_priorities=metadata.data_source.order,
                    separator=metadata.data_source.separator,
                    name=f"Value and priority order check for {DATA_SOURCE_COLUMN}.",
                )
            ],
        ),
        DATE_COLUMN: pa.Column(
            pa.DateTime,
            **default_non_required_kwargs(nullable=False),
            checks=[pa.Check(schema_checks.date_datetime_check, element_wise=True)],
        ),
        OPERATOR_COLUMN: pa.Column(
            pa.String,
            **default_non_required_kwargs(nullable=False),
            checks=[pa.Check.isin(metadata.operators)],
        ),
        SCALE_COLUMN: pa.Column(
            pa.String,
            **default_non_required_kwargs(nullable=False),
            checks=[
                prioritized_values_check(
                    named_priorities=metadata.scale.order,
                    separator=metadata.scale.separator,
                    name=f"Value and priority order check for {SCALE_COLUMN}.",
                )
            ],
        ),
        CERTAINTY_COLUMN: pa.Column(
            pa.String,
            **default_non_required_kwargs(nullable=False),
            checks=[pa.Check.isin(metadata.certainty)],
        ),
        LINEAMENT_ID_COLUMN: pa.Column(
            pa.String,
            **default_non_required_kwargs(nullable=False),
            checks=[
                pa.Check(
                    lambda raw_value: schema_checks.lineament_id_check(
                        raw_value=raw_value,
                        lineament_id_prefixes=metadata.lineament_id_prefixes,
                    ),
                    element_wise=True,
                    name=f"{LINEAMENT_ID_COLUMN} check.",
                )
            ],
            allow_duplicates=False,
        ),
    }
    return pa.DataFrameSchema(
        index=pa.Index(pa.Int),
        columns={
            "geometry": pa.Column(
                required=True,
            ),
            **trace_columns,
        },
    )



"""
General utilities.
"""
import json
import logging
from contextlib import suppress
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Sequence, Tuple, Type

import geopandas as gpd
import pandas as pd
import pandera as pa
from fractopo import general
from rich.table import Table
from rich.text import Text

from tracerepo import rules, trace_schema

GEOJSON_DRIVER = "GeoJSON"
EXPORT_DIR_PREFIX = "data-exported-"


class TraceTuple(NamedTuple):

    """
    Named tuple of traces and area paths and snap_threshold.
    """

    traces_path: Path
    area_path: Path
    validity: str
    snap_threshold: float = 0.001


@dataclass
class UpdateTuple:

    """
    Tuple with information of validity for trace-area-pair.
    """

    area_name: str
    update_values: Dict[rules.ColumnNames, str]
    traces_path: Path
    error: bool = False


def dataframe_column_to_python(
    dataframe: pd.DataFrame, column: str, python_type: Type[Any]
) -> List[Any]:
    """
    Get dataframe column as Python typed list.

    E.g.

    >>> df = pd.DataFrame({"data": [1.0, 52.2, 1.0]})
    >>> column = "data"
    >>> python_type = float
    >>> as_list = dataframe_column_to_python(df, column, python_type)
    >>> as_list
    [1.0, 52.2, 1.0]
    >>> hasattr(as_list[0], "item")
    False

    """
    # area name is used as index
    python_list: List[Any]
    if column == rules.ColumnNames.AREA.value:
        python_list = dataframe.index.to_list()
    else:
        python_list = dataframe[column].to_list()
    if not isinstance(python_list, list):
        raise TypeError("Expected Python list.")
    python_list_pythoned = [
        val.item() if hasattr(val, "item") else val for val in python_list
    ]
    python_list_checked = [
        val for val in python_list_pythoned if isinstance(val, python_type)
    ]
    if len(python_list) != len(python_list_checked):
        raise TypeError(
            f"Expected database to contain types convertible to Python {python_type}."
        )

    return python_list_checked


def compiled_path(
    thematic: str,
    geometry: str,
    scale: str,
    name: str,
    root: Path,
    data_root: str = rules.PathNames.DATA.value,
) -> Path:
    r"""
    Compile Path.

    E.g.

    >>> path = str(
    ...     compiled_path(
    ...         "inkoo", "traces", "drone_20m", "geta_20m_1_traces", root=Path(".")
    ...     )
    ... )
    >>> path.replace("\\", "/")
    'tracerepository_data/inkoo/traces/drone_20m/geta_20m_1_traces.geojson'

    """
    return root / data_root / thematic / geometry / scale / f"{name}.{rules.FILETYPE}"


def check_database_row_files(
    thematic: str,
    geometry: str,
    scale: str,
    name: str,
    root: Path,
):
    """
    Check that a row in database actually corresponds to trace and area files.
    """
    path = compiled_path(
        root=root, thematic=thematic, geometry=geometry, scale=scale, name=name
    )
    if not path.exists():
        raise FileNotFoundError(f"Expected {name} file to exist at {path}.")


def identify_geom_type(filename_stem: str) -> rules.ColumnNames:
    """
    Naively identify if file contains traces or areas based on filename.

    E.g.

    >>> identify_geom_type("geta_20m_1_traces")
    <ColumnNames.TRACES: 'traces'>

    """
    assert "." not in filename_stem
    if filename_stem.endswith("_area"):
        return rules.ColumnNames.AREA
    if filename_stem.endswith("_traces"):
        return rules.ColumnNames.TRACES
    raise ValueError(
        f"Expected filename_stem {filename_stem} to end in _area or _traces."
    )


def multi_string_filter(
    strings: Sequence[str], list_to_filter: Sequence[str]
) -> Sequence[bool]:
    """
    Filter list for any matching string in strings.

    E.g.

    >>> multi_string_filter(["geta"], ["geta", "heta"])
    [True, False]

    >>> multi_string_filter(["geta", "leta"], ["geta", "heta", "geta"])
    [True, False, True]

    >>> multi_string_filter([], ["geta", "heta", "geta"])
    [True, True, True]

    """
    if len(strings) == 0:
        return [True] * len(list_to_filter)

    string_filtered = [string_filter(string, list_to_filter) for string in strings]

    any_filtered = [any(vals) for vals in zip(*string_filtered)]

    assert len(any_filtered) == len(list_to_filter)

    return any_filtered


def string_filter(string: str, list_to_filter: Sequence[str]) -> Sequence[bool]:
    """
    Filter list for strings that contain string.
    """
    bools = [string in val for val in list_to_filter]
    return bools


def join_bools(*bool_sequences) -> Sequence[bool]:
    """
    Perform logical OR on sequences of bools of equal length.
    """
    joined = [all(vals) for vals in zip(*bool_sequences)]
    return joined


def query_result_tuple(
    thematic_val: str,
    scale_val: str,
    traces_val: str,
    area_val: str,
    validity_val: str,
    snap_threshold: float,
    tracerepository_path: Path,
) -> TraceTuple:
    """
    Compile TraceTuple with trace path, area path and snap_threshold.

    Some of paths might be None based on geometry_filter.
    """
    traces_path = compiled_path(
        root=tracerepository_path,
        thematic=thematic_val,
        scale=scale_val,
        name=traces_val,
        geometry=rules.ColumnNames.TRACES.value,
    )
    area_path = compiled_path(
        root=tracerepository_path,
        thematic=thematic_val,
        scale=scale_val,
        name=area_val,
        geometry=rules.ColumnNames.AREA.value,
    )

    return TraceTuple(
        traces_path=traces_path,
        area_path=area_path,
        snap_threshold=snap_threshold,
        validity=validity_val,
    )


def convert_sequence_columns(gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """
    Convert sequence (list/tuple) type columns to string.
    """
    gdf = gdf.copy()
    if gdf.empty:
        return gdf
    for column in gdf.columns.values:
        column_data = gdf[column]
        assert isinstance(column_data, pd.Series)
        first_val = column_data.values[0]
        if isinstance(first_val, (list, tuple)):
            logging.info(f"Converting {column} from {type(first_val)} to str.")
            gdf[column] = [str(tuple(item)) for item in column_data.values]
    return gdf


def write_geojson(gdf: gpd.GeoDataFrame, path: Path):
    """
    Write geodata as GeoJSON with 1 space delimitation.
    """
    as_json = gdf.to_json(indent=1)
    path.write_text(as_json)


def write_geodata(gdf: gpd.GeoDataFrame, path: Path, driver: str = GEOJSON_DRIVER):
    """
    Write geodata with driver.

    Default is GeoJSON.
    """
    if gdf.empty:
        # Handle empty GeoDataFrames
        path.write_text(gdf.to_json())
    else:
        gdf = convert_sequence_columns(gdf)

        gdf.to_file(path, driver=driver)

    if driver != GEOJSON_DRIVER:
        return

    # Format geojson with indent of 1
    read_json = path.read_text()
    loaded_json = json.loads(read_json)
    dumped_json = json.dumps(loaded_json, indent=1)
    path.write_text(dumped_json)


def rename_data_path(path: Path, rename_to: str) -> Path:
    """
    Rename the first directory of path to rename_to.

    >>> str(rename_data_path(Path("data/loviisa/traces/20m/file.txt"), "hey"))
    'hey/loviisa/traces/20m/file.txt'

    """
    splitter = "/" if "/" in str(path) else r"\\"

    return Path(f"{rename_to}/" + "/".join(str(path).split(splitter)[1:]))


def compile_export_dir(driver: str) -> str:
    """
    Compile directory name for data exporting.

    E.g.

    >>> compile_export_dir("ESRI Shapefile")
    'data-exported-ESRI-Shapefile'
    """
    dash_replaced_driver = driver.replace(" ", "-")
    return f"{EXPORT_DIR_PREFIX}{dash_replaced_driver}"


def remove_from_dict_if_in(key: str, dict_to_check: Dict[str, Path]):
    """
    Remove value from dict if it exists.

    E.g.

    >>> remove_from_dict_if_in("some_key", dict(some_key=Path(".")))
    {}

    >>> remove_from_dict_if_in("some_key", dict(other_key=Path(".")))
    {'other_key': PosixPath('.')}
    """
    with suppress(KeyError):
        dict_to_check.pop(key)
    return dict_to_check


def perform_pandera_check(
    traces: gpd.GeoDataFrame, metadata: rules.Metadata
) -> pd.DataFrame:
    """
    Validate the column data in ``traces`` ``GeoDataFrame``.
    """
    pandera_report: pd.DataFrame = pd.DataFrame()
    assert pandera_report.empty
    try:
        trace_schema.traces_schema(metadata=metadata).validate(traces, lazy=True)
    except pa.errors.SchemaErrors as exc:
        pandera_report = exc.failure_cases
        assert isinstance(pandera_report, pd.DataFrame)

    return pandera_report


def filename_friendly_datetime_string() -> str:
    """
    Get filename friendly datetime string.
    """
    return datetime.now().strftime("%Y%m%d_%H%M")


def report_pandera_errors(
    pandera_report: pd.DataFrame, report_directory: Path, area_name: str
) -> str:
    """
    Report pandera errors as html files saved to a reports directory.
    """
    report_directory.mkdir(exist_ok=True)
    current_time = filename_friendly_datetime_string()
    report_name = f"{area_name}_report_{current_time}.html"
    report_path = report_directory / report_name
    pandera_report.to_html(report_path)
    return f"Reported {area_name} traces pandera errors to {report_path}."


def otherwise_valid(update_tuple: UpdateTuple) -> bool:
    """
    Is dataset otherwise valid.
    """
    assert rules.ColumnNames.VALIDITY in update_tuple.update_values
    if (
        update_tuple.update_values[rules.ColumnNames.VALIDITY]
        == rules.ValidationResults.VALID.value
    ):
        return True
    return False


def pandera_reporting(
    update_tuple: UpdateTuple, metadata: rules.Metadata
) -> Tuple[Dict[rules.ColumnNames, str], pd.DataFrame]:
    """
    Check traces GeoDataFrame column data against schema and report if needed.
    """
    if update_tuple.update_values[rules.ColumnNames.VALIDITY] in (
        rules.ValidationResults.EMPTY.value,
        rules.ValidationResults.CRITICAL.value,
    ):
        return dict(), pd.DataFrame()
    # Read traces from disk.
    # (Alternative is to keep GeoDataFrame in memory from multiprocessing
    # but that is risky.)
    traces = general.read_geofile(update_tuple.traces_path)
    if traces.empty:
        logging.error(f"Empty traces uncaught by validation for {update_tuple}.")
        return dict(), pd.DataFrame()
    try:
        pandera_report = perform_pandera_check(traces, metadata=metadata)
    except Exception as exc:
        logging.error(
            f"GeoDataFrame validation critically failed with {update_tuple} traces.",
            exc_info=True,
        )
        pandera_report = pd.DataFrame(
            {"ERROR": ["Column validation critically failed...", str(exc)]}
        )
    if pandera_report.empty:
        return dict(), pd.DataFrame()

    if otherwise_valid(update_tuple=update_tuple):
        # If the dataset is otherwise marked valid mark it as unfit due
        # to pandera schema error
        update_values = {
            rules.ColumnNames.VALIDITY: rules.ValidationResults.UNFIT.value
        }
        return update_values, pandera_report
    return dict(), pandera_report


def create_validation_table(
    invalids: List[TraceTuple], validity_changes: Optional[List[Text]] = None
) -> Table:
    """
    Generate a rich Table from invalids.
    """
    if validity_changes is not None:
        assert len(validity_changes) == len(invalids)
        title = "Validation Results"
    else:
        title = "Validation Targets"
    table = Table(title=title)
    table.add_column("Traces", header_style="bold")
    table.add_column("Area", header_style="bold")
    table.add_column("Snap Threshold", header_style="bold", style="bold blue")
    # table.add_column("Validity", header_style="bold", style="bold blue")
    table.add_column("Validity")

    for idx, trace_tuple in enumerate(invalids):
        validity = (
            enrich_color_validity_value(trace_tuple.validity)
            if validity_changes is None
            else validity_changes[idx]
        )
        table.add_row(
            trace_tuple.traces_path.name,
            trace_tuple.area_path.name,
            str(trace_tuple.snap_threshold),
            validity,
        )
    return table


def enrich_color_validity_value(validity_value: str) -> Text:
    """
    Wrap validity value with approppriate color.
    """
    try:
        color = rules.ValidationResults.color_dict()[validity_value]
    except KeyError:
        logging.error(
            "validity_value was not found in rules.ValidationResults.color_dict()",
            extra=dict(validity_value=validity_value),
            exc_info=True,
        )
        color = "blue"
    return Text(text=validity_value, style=color)


def create_validation_results_table(
    invalids: List[TraceTuple], update_tuples: List[UpdateTuple]
) -> Table:
    """
    Create table for validation results.
    """
    assert len(invalids) == len(update_tuples)
    new_validities = [
        update_tuple.update_values[rules.ColumnNames.VALIDITY]
        for update_tuple in update_tuples
    ]
    old_validities = [invalid.validity for invalid in invalids]

    validity_changes = [
        Text.assemble(
            enrich_color_validity_value(old), " -> ", enrich_color_validity_value(new)
        )
        for old, new in zip(old_validities, new_validities)
    ]

    table = create_validation_table(
        invalids=invalids, validity_changes=validity_changes
    )
    return table
